@@ -3,20 +3,26 @@ extern crate proc_macro;
 use syn::parse::Parser;
 use syn::punctuated::Punctuated;
 use syn::spanned::Spanned;
-use syn::token::Comma;
 
 #[proc_macro_derive(Asn1Read, attributes(explicit, implicit, default, defined_by))]
 pub fn derive_asn1_read(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
     let input = syn::parse_macro_input!(input as syn::DeriveInput);
 
     let name = input.ident;
-    let (impl_lifetimes, ty_lifetimes, lifetime_name) = add_lifetime_if_none(input.generics);
+    let (_, ty_generics, _) = input.generics.split_for_impl();
+    let mut generics = input.generics.clone();
+    let lifetime_name = add_lifetime_if_none(&mut generics);
+    add_bounds(
+        &mut generics,
+        syn::parse_quote!(asn1::Asn1Readable<#lifetime_name>),
+    );
+    let (impl_generics, _, where_clause) = generics.split_for_impl();
 
     let expanded = match input.data {
         syn::Data::Struct(data) => {
             let read_block = generate_struct_read_block(&name, &data);
             quote::quote! {
-                impl<#impl_lifetimes> asn1::SimpleAsn1Readable<#lifetime_name> for #name<#ty_lifetimes> {
+                impl #impl_generics asn1::SimpleAsn1Readable<#lifetime_name> for #name #ty_generics #where_clause {
                     const TAG: asn1::Tag = <asn1::Sequence as asn1::SimpleAsn1Readable>::TAG;
                     fn parse_data(data: &#lifetime_name [u8]) -> asn1::ParseResult<Self> {
                         asn1::parse(data, |p| { #read_block })
@@ -27,7 +33,7 @@ pub fn derive_asn1_read(input: proc_macro::TokenStream) -> proc_macro::TokenStre
         syn::Data::Enum(data) => {
             let (read_block, can_parse_block) = generate_enum_read_block(&name, &data);
             quote::quote! {
-                impl<#impl_lifetimes> asn1::Asn1Readable<#lifetime_name> for #name<#ty_lifetimes> {
+                impl #impl_generics asn1::Asn1Readable<#lifetime_name> for #name #ty_generics #where_clause {
                     fn parse(parser: &mut asn1::Parser<#lifetime_name>) -> asn1::ParseResult<Self> {
                         let tlv = parser.read_element::<asn1::Tlv>()?;
                         #read_block
@@ -49,16 +55,17 @@ pub fn derive_asn1_read(input: proc_macro::TokenStream) -> proc_macro::TokenStre
 
 #[proc_macro_derive(Asn1Write, attributes(explicit, implicit, default, defined_by))]
 pub fn derive_asn1_write(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
-    let input = syn::parse_macro_input!(input as syn::DeriveInput);
+    let mut input = syn::parse_macro_input!(input as syn::DeriveInput);
 
     let name = input.ident;
-    let lifetimes = find_lifetimes(input.generics);
+    add_bounds(&mut input.generics, syn::parse_quote!(asn1::Asn1Writable));
+    let (impl_generics, ty_generics, where_clause) = input.generics.split_for_impl();
 
     let expanded = match input.data {
         syn::Data::Struct(data) => {
             let write_block = generate_struct_write_block(&data);
             quote::quote! {
-                impl<#lifetimes> asn1::SimpleAsn1Writable for #name<#lifetimes> {
+                impl #impl_generics asn1::SimpleAsn1Writable for #name #ty_generics #where_clause {
                     const TAG: asn1::Tag = <asn1::SequenceWriter as asn1::SimpleAsn1Writable>::TAG;
                     fn write_data(&self, dest: &mut asn1::WriteBuf) -> asn1::WriteResult {
                         #write_block
@@ -71,7 +78,7 @@ pub fn derive_asn1_write(input: proc_macro::TokenStream) -> proc_macro::TokenStr
         syn::Data::Enum(data) => {
             let write_block = generate_enum_write_block(&name, &data);
             quote::quote! {
-                impl<#lifetimes> asn1::Asn1Writable for #name<#lifetimes> {
+                impl #impl_generics asn1::Asn1Writable for #name #ty_generics #where_clause {
                     fn write(&self, w: &mut asn1::Writer) -> asn1::WriteResult {
                         #write_block
                     }
@@ -123,7 +130,10 @@ pub fn derive_asn1_defined_by_read(input: proc_macro::TokenStream) -> proc_macro
     let input = syn::parse_macro_input!(input as syn::DeriveInput);
 
     let name = input.ident;
-    let (impl_lifetimes, ty_lifetimes, lifetime_name) = add_lifetime_if_none(input.generics);
+    let (_, ty_generics, where_clause) = input.generics.split_for_impl();
+    let mut generics = input.generics.clone();
+    let lifetime_name = add_lifetime_if_none(&mut generics);
+    let (impl_generics, _, _) = generics.split_for_impl();
 
     let mut read_block = vec![];
     let mut default_ident = None;
@@ -166,7 +176,7 @@ pub fn derive_asn1_defined_by_read(input: proc_macro::TokenStream) -> proc_macro
         }
     };
     proc_macro::TokenStream::from(quote::quote! {
-        impl<#impl_lifetimes> asn1::Asn1DefinedByReadable<#lifetime_name, asn1::ObjectIdentifier> for #name<#ty_lifetimes> {
+        impl #impl_generics asn1::Asn1DefinedByReadable<#lifetime_name, asn1::ObjectIdentifier> for #name #ty_generics #where_clause {
             fn parse(item: asn1::ObjectIdentifier, parser: &mut asn1::Parser<#lifetime_name>) -> asn1::ParseResult<Self> {
                 #(#read_block)*
 
@@ -181,7 +191,7 @@ pub fn derive_asn1_defined_by_write(input: proc_macro::TokenStream) -> proc_macr
     let input = syn::parse_macro_input!(input as syn::DeriveInput);
 
     let name = input.ident;
-    let lifetimes = find_lifetimes(input.generics);
+    let (impl_generics, ty_generics, where_clause) = input.generics.split_for_impl();
 
     let mut write_blocks = vec![];
     let mut item_blocks = vec![];
@@ -222,7 +232,7 @@ pub fn derive_asn1_defined_by_write(input: proc_macro::TokenStream) -> proc_macr
     }
 
     proc_macro::TokenStream::from(quote::quote! {
-        impl<#lifetimes> asn1::Asn1DefinedByWritable<asn1::ObjectIdentifier> for #name<#lifetimes> {
+        impl #impl_generics asn1::Asn1DefinedByWritable<asn1::ObjectIdentifier> for #name #ty_generics #where_clause {
             fn item(&self) -> &asn1::ObjectIdentifier {
                 match self {
                     #(#item_blocks)*
@@ -238,33 +248,24 @@ pub fn derive_asn1_defined_by_write(input: proc_macro::TokenStream) -> proc_macr
     })
 }
 
-fn find_lifetimes(mut generics: syn::Generics) -> Punctuated<syn::Lifetime, Comma> {
-    let mut lifetimes = Punctuated::new();
-    for param in &mut generics.params {
-        if let syn::GenericParam::Lifetime(lifetime_def) = param {
-            lifetimes.push(lifetime_def.lifetime.clone());
-        }
-    }
+fn add_lifetime_if_none(generics: &mut syn::Generics) -> syn::Lifetime {
+    if generics.lifetimes().next().is_none() {
+        generics
+            .params
+            .push(syn::GenericParam::Lifetime(syn::LifetimeParam::new(
+                syn::Lifetime::new("'a", proc_macro2::Span::call_site()),
+            )));
+    };
 
-    lifetimes
+    generics.lifetimes().next().unwrap().lifetime.clone()
 }
 
-fn add_lifetime_if_none(
-    generics: syn::Generics,
-) -> (
-    Punctuated<syn::Lifetime, Comma>,
-    Punctuated<syn::Lifetime, Comma>,
-    syn::Lifetime,
-) {
-    let mut impl_lifetimes = find_lifetimes(generics);
-    let ty_lifetimes = impl_lifetimes.clone();
-    let lifetime = impl_lifetimes.first().cloned().unwrap_or_else(|| {
-        let lifetime = syn::Lifetime::new("'a", proc_macro2::Span::call_site());
-        impl_lifetimes.push(lifetime.clone());
-        lifetime
-    });
-
-    (impl_lifetimes, ty_lifetimes, lifetime)
+fn add_bounds(generics: &mut syn::Generics, bound: syn::TypeParamBound) {
+    for param in &mut generics.params {
+        if let syn::GenericParam::Type(ref mut type_param) = param {
+            type_param.bounds.push(bound.clone());
+        }
+    }
 }
 
 enum OpType {
